from dataclasses import dataclass
from math import log

from ...rng import RNG
from ...stats.distributions import Distribution
from .. import Internal, Node, Operator, Proposal, Tree
from ._util import assert_factor, sample_range


@dataclass(slots=True)
class TreeScale(Operator):
    """Scales the age of the entire tree

    This parameter is analogous to BEAST2's `ScaleOperator` when it's used on a
    tree.  It will scale all internal nodes by a random scale which is randomly
    picked depending on `factor` and `distribution`.
    """

    tree: Tree
    """The tree to scale."""
    factor: float
    """
    The scaling ratio will be sampled from `(factor, 1 / factor)`.  So, the
    factor must be between 0 and 1 and the smaller it is the larger the steps
    will be.
    """
    distribution: Distribution
    """Distribution from which the scale is sampled."""
    rng: RNG
    weight: float = 1

    def __post_init__(self):
        assert_factor(self)

    def propose(self) -> Proposal:
        tree = self.tree
        rng = self.rng
        root = tree.root

        low, high = self.factor, 1 / self.factor
        scale = sample_range(low, high, self.distribution, rng)

        try:
            tree.scale(scale)
        except:
            return Proposal.Reject()

        ratio = log(scale) * (tree.num_internals - 2)
        return Proposal.Hastings(ratio)



from dataclasses import dataclass
from math import log

from ...rng import RNG
from .. import Internal, Node, Operator, Proposal, Tree
from ._util import assert_two_internals


@dataclass(slots=True)
class NarrowExchange(Operator):
    """Exchanges the parents of two neighbouring nodes

    This operator is analogous to BEAST2's `Exchange` operator with `isNarrow`
    set to true.  It finds a grandparent (internal node both of whose children
    are also internal) with two kids: `parent` and `uncle` (uncle is younger
    than the parent).  And one of the children of `parent` is swapped with
    `uncle`.
    """

    tree: Tree
    rng: RNG
    weight: float = 1

    def __post_init__(self):
        assert_two_internals(self)

    def propose(self) -> Proposal:
        tree = self.tree

        num_grandparents_before = tree.num_grandparents()
        if num_grandparents_before == 0:
            # no grandparents to pick `grandparent` from
            return Proposal.Reject()

        while True:
            grandparent = tree.random_internal(self.rng)
            if tree.is_grandparent(grandparent):
                break

        left, right = tree.children_of(grandparent)
        if tree.height_of(left) > tree.height_of(right):
            parent, uncle = left, right
        elif tree.height_of(right) > tree.height_of(left):
            parent, uncle = right, left
        else:
            return Proposal.Reject()

        # guaranteed because `grandparent` is a grandparent
        assert isinstance(parent, Internal)
        assert isinstance(uncle, Internal)

        before = int(tree.is_grandparent(parent)) + int(tree.is_grandparent(uncle))

        if self.rng.random_bool(0.5):
            child = tree.children_of(parent)[0]
        else:
            child = tree.children_of(parent)[1]

        tree.swap_parents(uncle, child)

        after = int(tree.is_grandparent(parent)) + int(tree.is_grandparent(uncle))
        num_grandparents_after = num_grandparents_before - before + after
        ratio = log(num_grandparents_before / num_grandparents_after)

        return Proposal.Hastings(ratio)


@dataclass(slots=True)
class WideExchange(Operator):
    """Exchanges the parent of two random nodes

    This operator is analogous to BEAST2's `Exchange` operator with `isNarrow`
    set to false.  It picks two random nodes in the tree (they could be either
    leaves or internals) and swaps their parents.

    If a randomly selected move is impossible (a parent would be younger than
    its child) the operator aborts with `Proposal.Reject`.
    """

    tree: Tree
    rng: RNG
    weight: float = 1

    def propose(self) -> Proposal:
        tree = self.tree

        root = tree.root

        i = tree.random_node(self.rng)
        while i == root:
            i = tree.random_node(self.rng)

        j = None
        while j is None or j == i or j == root:
            j = tree.random_node(self.rng)
        assert j is not None

        i_parent = tree.parent_of(i)
        if i_parent is None:
            return Proposal.Reject()
        j_parent = tree.parent_of(j)
        if j_parent is None:
            return Proposal.Reject()

        # Abort if j and i are parent-child or if one of the parents would be
        # younger than its new child or if the two selected nodes.
        if (
            j != i_parent
            and i != j_parent
            and tree.height_of(j) < tree.height_of(i_parent)
            and tree.height_of(i) < tree.height_of(j_parent)
        ):
            tree.swap_parents(i, j)

            return Proposal.Hastings(0.0)
        else:
            return Proposal.Reject()


@dataclass(slots=True)
class BeastNarrowExchange(Operator):
    """
    Narrow exchange operator compatible with BEASTs `narrowExchange`
    """

    tree: Tree
    rng: RNG
    weight: float = 1

    def propose(self):
        rng = self.rng
        tree = self.tree
        root = tree.root

        node = tree.random_node(rng)
        while node == root or tree.parent_of(node) == root:
            node = tree.random_node(rng)

        parent = tree.parent_of(node)
        assert parent is not None
        grandparent = tree.parent_of(parent)
        assert grandparent is not None
        uncle = tree.other_child(grandparent, parent)

        if tree.height_of(uncle) < tree.height_of(parent):
            tree.swap_parents(node, uncle)
            return Proposal.Hastings(0.0)
        else:
            return Proposal.Reject()


@dataclass(slots=True)
class BeastWideExchange(Operator):
    """
    Wide exchange operator compatible with BEASTs `wideExchange`
    """

    tree: Tree
    rng: RNG
    weight: float = 1

    def propose(self):
        rng = self.rng
        tree = self.tree
        root = tree.root

        node_a = tree.random_node(rng)
        while node_a == root:
            node_a = tree.random_node(rng)

        node_b = tree.random_node(rng)
        while node_b == root or node_b == node_a:
            node_b = tree.random_node(rng)

        node_a_parent = tree.parent_of(node_a)
        node_b_parent = tree.parent_of(node_b)
        assert node_a_parent is not None
        assert node_b_parent is not None

        if (
            node_a_parent != node_b_parent
            and node_a != node_b_parent
            and node_b != node_a_parent
            and tree.height_of(node_a) < tree.height_of(node_b_parent)
            and tree.height_of(node_b) < tree.height_of(node_a_parent)
        ):
            tree.swap_parents(node_a, node_b)
            return Proposal.Hastings(0.0)
        else:
            return Proposal.Reject()



"""Computational biology toolkit for Python powered by Rust.

- `b3`: Bayesian phylogenetic analysis engine, analogous to BEAST2.
- `data`: biological data classes, currently only include DNA.
- `io`: bioinformatics file formats parsers.
- `rng`: random number generator used by `b3` and `stats`.
- `stats`: statistical functions.
"""



"""Utilities for basic statistics.

This library aims to mirror the functionality of `scipy.stats`.  The reason the
latter could not be used is because SciPy is optimized for batch processing,
which causes it to be quite slow for singular calls, like creating a new
distribution and calculating a PDF of one value.  `b3` operators and priors
have this exact usage pattern[^params], so `stats` was created as a library
which performs reasonably well on such workloads.


[^params]:
    Since each MCMC step depends on the last one priors have to be recalculated
    on each step.  And operators also only propose one move at a time.
    Additionally, since parameters can change, it's easier to recreate
    distribution classes on each step, but SciPy ones were definitely not
    intended for that.
"""



from math import exp, inf, isfinite, log

from ...rng import RNG
from .. import Internal, Node, Tree


# x must be in [0, inf)
def interval_to_range(ratio: float, low: float, high: float):
    return low + (high - low) / (ratio + 1)


def _is_on_range(distribution) -> bool:
    return isfinite(distribution.lower) and isfinite(distribution.upper)


def _sample_rescale(low, high, distribution, rng: RNG):
    x = distribution.sample(rng)
    ratio = (x - distribution.lower) / (distribution.upper - distribution.lower)
    # `ratio` is in [0, 1]
    return low + (high - low) * ratio


def sample_range(low: float, high: float, distribution, rng: RNG) -> int | float:
    if _is_on_range(distribution):
        return _sample_rescale(low, high, distribution, rng)

    x = distribution.sample(rng)

    # if the distribution is full-line rescale it to positive numbers only
    if distribution.lower == -inf:
        x = exp(x)

    # fold lines and half-open intervals into a range
    new_point = interval_to_range(x, low, high)
    return new_point


def family(tree: Tree, rng: RNG) -> tuple[Node, Node, Internal, Internal | None]:
    root = tree.root

    node = tree.random_node(rng)
    while node == root:
        node = tree.random_node(rng)

    parent = tree.parent_of(node)
    assert parent is not None  # checked in the loop

    grandparent = tree.parent_of(parent)

    sibling = tree.other_child(parent, node)

    return node, sibling, parent, grandparent


def assert_two_internals(operator) -> None:
    if operator.tree.num_internals < 2:
        raise ValueError(
            f"`{operator.__class__.__name__}` requires the tree to have at least 2 internal nodes"
        )


def assert_factor(operator) -> None:
    if not (0.0 < operator.factor < 1.0):
        raise ValueError(f"Factor must be in range (0, 1), got {operator.factor}")



"""
Felsenstein's tree likelihood calculators.
"""

from dataclasses import dataclass
from typing import Protocol

from .._aspartik_rust_impl._b3_rust_impl import (
    CPU4Likelihood as CPU4Likelihood,
    CUDALikelihood as CUDALikelihood,
    Thread4Likelihood as Thread4Likelihood,
)
from . import Stateful
from .parameters import Weights


class Likelihood(Stateful, Protocol):
    """
    Tree likelihood calculator

    This object calculates the likelihood of a tree given the sequence data
    using Felsenstein's tree pruning algorithm.

    There are several implementations, each with its own options.
    """

    def propose(self) -> None:
        """
        Fetches the current model state and starts the calculations.

        Each implementations are responsible for pulling the state.

        Depending on the implementation, `propose` might start the calculation
        in parallel and return.  In this case `likelihood` will block on it.
        """
        ...

    def likelihood(self) -> float:
        """
        Returns the tree likelihood calculated.

        If `propose` was called during this epoch, this must the likelihood as
        a result of the proposal.  If `accept` or `reject` was called last,
        this method must return the last accepted likelihood.
        """
        ...


@dataclass(slots=True)
class CompoundLikelihood(Likelihood):
    """
    Combines several likelihoods.

    `CompoundLikelihood` doesn't use a thread pool, but depending on the
    `Likelihood` implementers, calculations might be done in parallel.
    """

    likelihoods: list[Likelihood]

    def propose(self) -> None:
        for likelihood in self.likelihoods:
            likelihood.propose()

    def likelihood(self) -> float:
        return sum(l.likelihood() for l in self.likelihoods)

    def accept(self) -> None:
        for likelihood in self.likelihoods:
            likelihood.accept()

    def reject(self) -> None:
        for likelihood in self.likelihoods:
            likelihood.reject()


@dataclass(slots=True)
class WeightedLikelihood(Likelihood):
    """
    Combines several likelihoods with dynamic weights.

    The formula is `Σ l_i * w_i`.  It is inspired [GHOST] and can be used to
    analyse several sequence evolution models.  Much like `CompoundLikelihood`,
    most likelihood implementations will be

    [GHOST]: https://doi.org/10.1093/sysbio/syz051
    """

    likelihoods: list[Likelihood]
    weights: Weights

    def propose(self) -> None:
        for likelihood in self.likelihoods:
            likelihood.propose()

    def likelihood(self) -> float:
        out = 0
        for likelihood, weight in zip(self.likelihoods, self.weights):
            out += likelihood.likelihood() * weight
        return out

    def accept(self) -> None:
        for likelihood in self.likelihoods:
            likelihood.accept()

    def reject(self) -> None:
        for likelihood in self.likelihoods:
            likelihood.reject()



from abc import abstractmethod
from typing import Protocol


class Comparable(Protocol):
    @abstractmethod
    def __lt__(self, other) -> bool: ...
    @abstractmethod
    def __le__(self, other) -> bool: ...
    @abstractmethod
    def __eq__(self, other) -> bool: ...
    @abstractmethod
    def __ne__(self, other) -> bool: ...
    @abstractmethod
    def __gt__(self, other) -> bool: ...
    @abstractmethod
    def __ge__(self, other) -> bool: ...


class SimpleSequence[T](Protocol):
    @abstractmethod
    def __len__(self) -> int: ...
    @abstractmethod
    def __getitem__(self, i: int) -> T: ...
    @abstractmethod
    def __setitem__(self, i: int, value: T): ...



from dataclasses import dataclass

from ...rng import RNG
from ...stats.distributions import Distribution
from .. import Operator, Proposal, Tree
from ._util import assert_two_internals, sample_range


@dataclass(slots=True)
class NodeSlide(Operator):
    """Slides the age of a random internal node between its parent and children

    This operator is similar to BEAST2's `EpochFlexOperator`: it will only
    affect the age of the selected node without altering the tree topology (a
    node cannot slide above its parent).
    """

    tree: Tree
    """The tree to edit."""
    distribution: Distribution
    """
    The distribution which will sample the new node height on the interval
    between its parent and the closest child.
    """
    rng: RNG
    weight: float = 1

    def __post_init__(self):
        assert_two_internals(self)

    def propose(self) -> Proposal:
        tree = self.tree

        node, parent = tree.random_nonroot_internal(self.rng)
        left, right = tree.children_of(node)

        oldest = tree.height_of(parent)
        youngest = max(tree.height_of(left), tree.height_of(right))

        new_height = sample_range(youngest, oldest, self.distribution, self.rng)

        tree.set_height(node, new_height)

        return Proposal.Hastings(0.0)



import math
from dataclasses import dataclass, field
from typing import Literal

from ...rng import RNG
from ...stats.distributions import Uniform
from .. import Operator, Proposal, Tree
from ..parameters import Real
from ._util import sample_range


@dataclass(slots=True)
class RandomWalk(Operator):
    """
    Adds or subtracts values to a parameter irrespective of its value

    The delta is sampled uniformly from `[0, window)` and is negated half of
    the time.

    This parameter doesn't have bounds.  Use the `Bound` prior to restrict
    parameter values.
    """

    param: Real
    window: float
    rng: RNG
    lower: float = 0
    upper: float = math.inf
    boundary: Literal["reflect"] = "reflect"
    weight: float = 1

    _dist: Uniform = field(default_factory=lambda: Uniform(0, 1), init=False)

    def propose(self) -> Proposal:
        lower, upper = self.lower, self.upper
        rng = self.rng
        param = self.param

        diff = sample_range(0, self.window, self._dist, rng)
        if rng.random_bool():
            diff *= -1

        new_value = float(param) + diff

        if new_value < self.lower:
            match self.boundary:
                case "reflect":
                    if self.upper == math.inf:
                        new_value = lower + (lower - new_value)
                    else:
                        # TODO: ping-pong reflection
                        pass

        if new_value > self.upper:
            # TODO
            if self.lower == math.inf:
                new_value = upper + (upper - new_value)
            else:
                pass

        self.param.set(new_value)

        return Proposal.Hastings(0.0)



from __future__ import annotations

from collections.abc import Sequence
from typing import Optional, Self

class DNANucleotide:
    Adenine: Self
    Cytosine: Self
    Guanine: Self
    Thymine: Self

    Weak: Self
    Strong: Self
    Amino: Self
    Ketone: Self
    Purine: Self
    Pyrimidine: Self

    NotAdenine: Self
    NotCytosine: Self
    NotGuanine: Self
    NotThymine: Self

    Any: Self
    Gap: Self

    def __init__(self, char: str): ...
    def __contains__(self, other: Self) -> bool: ...
    def complement(self) -> Self: ...

class DNASeq:
    def __init__(self, sequence: str | None): ...
    def __getitem__(self, index: int) -> DNANucleotide: ...
    def __len__(self) -> int: ...
    def complement(self) -> DNASeq: ...
    def reverse_complement(self) -> DNASeq: ...

class DNARecord:
    def __init__(self, name: str, sequence: DNASeq): ...
    @property
    def sequence(self) -> DNASeq: ...
    @property
    def description(self) -> str: ...
    @property
    def id(self) -> str: ...

class Phred:
    def __init__(self, ch: str): ...
    def accuracy(self) -> float: ...
    def probability_incorrect(self) -> float: ...

class MSA:
    @classmethod
    def from_fasta(cls, fasta: Sequence[DNARecord]) -> MSA:
        """
        Create an alignment from FASTA records

        All records must have the same length.
        """

    @property
    def num_sites(self) -> int:
        """
        Number of sites in the alignment

        Some alignments might be partial, that is, they might reference fewer
        sites than their backing buffer has.  This can happen after slicing or
        some methods.
        """

    @property
    def num_sequences(self) -> int:
        """Number of sequences in the alignment"""

    def sequence_name(self, index: int) -> str:
        """The name (or id) of the `index`'th sequence"""

    def sequence_names(self) -> list[str]:
        """A list of all sequence names in the alignment"""

    def sequence(self, index: int) -> DNASeq:
        """The `index`'th sequence in the alignment"""

    def base_frequencies(self) -> tuple[float, float, float, float]:
        """
        Base frequencies of Adenine, Cytosine, Guanine, and Thymine, in that
        order
        """

    from typing import Optional

class Tree:
    def __init__(self, newick: Optional[str] = None): ...
    def __str__(self) -> str: ...



from dataclasses import dataclass
from typing import Protocol, SupportsFloat


class Clock(Protocol):
    """
    A clock model

    Currently this class only supports strict clock models.
    """

    def get_rate(self) -> float:
        """
        The uniform clock rate for all edges in the tree
        """
        ...


@dataclass(slots=True)
class StrictClock(Clock):
    """Clock model which just returns a parameter"""

    mu: SupportsFloat

    def get_rate(self) -> float:
        return float(self.mu)



from dataclasses import dataclass
from math import inf
from typing import Protocol

from ...utils.typing import Comparable
from .. import Prior
from ..parameters import Real


@dataclass(slots=True)
class Bound(Prior):
    """Puts limits on the value of a parameter

    This prior serves the same purpose as the `lower` and `upper` attributes on
    BEAST parameters.  It will return `1` if all dimensions of the parameter
    lie within `[lower, upper)` or cancel the proposal by returning negative
    infinity otherwise.

    Due to how the internals of `b3` work, these priors should be first in the
    `priors` list in `run`, to avoid calculating other priors and likelihood if
    the bounds aren't satisfied.
    """

    param: Comparable
    """The parameter to be constrained."""
    lower: float = 0
    """Minimum possible value of the parameter, inclusive."""
    upper: float = inf
    """Maximum value of the parameter, exclusive (strictly compared)."""

    def probability(self) -> float:
        if not (self.lower <= self.param < self.upper):
            return -inf
        else:
            return 0



from dataclasses import dataclass

from ...rng import RNG
from ...stats.distributions import Distribution
from .. import Operator, Proposal, Tree
from ._util import assert_two_internals, family


@dataclass(slots=True)
class SubtreePruneRegraft(Operator):
    """
    Fixed height subtree and regraft move.

    This operator was described in [Hoehna et al 2008][l], section 3.2.7.  The
    move selects a random node `i` and its parent `i_parent`.  It then selects
    a random edge whose height overlaps with the height of `i_parent`.
    `i_parent` is spliced into the middle of this edge.

    [l]: https://alexeidrummond.org/assets/publications/2008-hoehna-evalution.pdf
    """

    tree: Tree
    """The tree to edit."""
    rng: RNG
    weight: float = 1

    def __post_init__(self):
        assert_two_internals(self)

    def propose(self) -> Proposal:
        rng = self.rng
        tree = self.tree
        root = tree.root

        node = tree.random_node(rng)
        parent = tree.parent_of(node)
        while node == root or parent == root:
            node = tree.random_node(rng)
            parent = tree.parent_of(node)

        assert parent is not None
        grandparent = tree.parent_of(parent)
        assert grandparent is not None

        sibling = tree.other_child(parent, node)

        parent_height = tree.height_of(parent)

        edge = tree.random_intersecting_edge(parent_height, rng)

        if edge is None:
            return Proposal.Reject()

        other, other_parent = tree.edge_nodes(edge)

        # Tree update
        grandparent_to_parent = tree.edge_index(parent)
        parent_to_sibling = tree.edge_index(sibling)
        other_parent_to_other = tree.edge_index(other)

        tree.update_edge(grandparent_to_parent, sibling)
        tree.update_edge(other_parent_to_other, parent)
        tree.update_edge(parent_to_sibling, other)

        tree.validate()

        return Proposal.Hastings(0.0)



from .. import MCMC


def print_operator_stats(mcmc: MCMC) -> None:
    print(f"{'Operator': <20}{'%accepts': >20}{'%aborts': >20}{'%prior rejects': >20}")
    print("-" * 80)

    for operator, results, _, _ in mcmc.operator_statistics:
        (
            unconditional_accepts,
            unconditional_rejects,
            prior_rejects,
            accepts,
            rejects,
        ) = results
        total = sum(results)
        accepts_share = accepts / total
        aborts_share = unconditional_rejects / total
        prior_share = prior_rejects / total
        print(
            f"{type(operator).__name__: <20}{accepts_share: >20.0%}{aborts_share: >20.0%}{prior_share: >20.0%}"
        )


def print_operator_timings(mcmc: MCMC) -> None:
    print(
        f"{'Operator': <20}{'propose avg μs': >20}{'likelihood avg μs': >20}{'total μs': >20}"
    )
    print("-" * 80)

    for operator, results, propose, likelihood in mcmc.operator_statistics:
        (
            unconditional_accepts,
            unconditional_rejects,
            prior_rejects,
            accepts,
            rejects,
        ) = results
        name = type(operator).__name__
        propose = (propose / (prior_rejects + accepts + rejects)).microseconds
        likelihood = (likelihood / (accepts + rejects)).microseconds
        total = propose + likelihood

        print(f"{name: <20}{propose: >20.0f}{likelihood: >20.0f}{total: >20.0f}")



from dataclasses import dataclass
from typing import Protocol

from ...rng import RNG
from ...utils.typing import SimpleSequence
from .. import Operator, Proposal
from ..parameters import Real


@dataclass(slots=True)
class DeltaExchange(Operator):
    """Scales a multidimensional parameter without changing its sum

    This operator is analogous to BEAST's `DeltaExchangeOperator`.  It picks
    two random dimensions from a parameter, samples a random delta from
    `distribution`, and increments one of them by delta and decrements the
    other one.
    """

    param: SimpleSequence[float]
    """
    The multidimensional parameter to edit.  Two random dimensions ones will be
    changed for each proposal.
    """
    factor: float
    """
    The move size is a random value between 0 and 1 multiplied by `factor`.
    """
    rng: RNG
    weight: float = 1

    def __post_init__(self):
        if len(self.param) <= 1:
            raise ValueError(f"`param` must have at least two dimensions")

    def propose(self) -> Proposal:
        rng = self.rng

        delta = rng.random_float() * self.factor

        dim_1 = rng.random_int(0, len(self.param))
        dim_2 = dim_1
        while dim_2 == dim_1:
            dim_1 = rng.random_int(0, len(self.param))

        self.param[dim_1] -= delta
        self.param[dim_2] += delta

        # The move is symmetrical, so the Hastings ratio is 0
        return Proposal.Hastings(0)



from ..._aspartik_rust_impl import _b3_rust_impl
from ._bound import Bound as Bound
from ._distribution import Distribution as Distribution

Yule = _b3_rust_impl.Yule
ConstantPopulation = _b3_rust_impl.ConstantPopulation
ExponentialGrowth = _b3_rust_impl.ExponentialGrowth
Monophyly = _b3_rust_impl.Monophyly


__all__ = [
    # Python
    "Bound",
    "Distribution",
    # Rust
    "ConstantPopulation",
    "Yule",
]



from dataclasses import dataclass
from math import log
from typing import Literal

from ...rng import RNG
from ...stats.distributions import Distribution
from .. import Operator, Proposal
from ..parameters import Real
from ._util import assert_factor, sample_range


@dataclass(slots=True)
class ParamScale(Operator):
    """Scales one parameter

    This operator is analogous to BEAST2's `ScaleOperator`, except it only
    works for parameters.

    Note that this operator doesn't have the upper/lower bounds BEAST2's analog
    has.  Instead the `Bound` prior should be used to put limits on the
    parameter values.
    """

    param: Real
    """The parameter to scale."""
    factor: float
    """
    The scale ratio will be sampled from `(factor, 1 / factor)`.  So, the
    smaller the factor, the larger the moves proposed by this operator are.
    Also, this means that `factor` must be within `(0, 1)`.
    """
    distribution: Distribution
    """The distribution from which to sample the scaling factor."""
    rng: RNG
    weight: float = 1

    def __post_init__(self):
        assert_factor(self)

    def propose(self) -> Proposal:
        low, high = self.factor, 1 / self.factor
        scale = sample_range(low, high, self.distribution, self.rng)

        self.param.scale(scale)

        ratio = -log(scale)
        return Proposal.Hastings(ratio)



from typing import Protocol, runtime_checkable

from .._aspartik_rust_impl._b3_rust_impl import (
    MCMC as MCMC,
    Internal as Internal,
    Leaf as Leaf,
    Proposal as Proposal,
    Tree as Tree,
)

type Node = Leaf | Internal
"""Any node of the phylogenetic tree

Used for type hints in places where there isn't a need to distinguish between
internal and leaf nodes.
"""


@runtime_checkable
class Prior(Protocol):
    """
    Interface which describes all prior distributions

    A `Prior` object will be queried by `MCMC` on each step after the operator
    edits the state to get the prior probability of the new state.  It's the
    responsibility of the prior to track the stateful objects it's interested
    in, so they will typically be passed in the constructor.

    Variable priors, coalescents, birth-death models, and all other
    non-likelihood models implement this protocol.
    """

    def probability(self) -> float:
        """Calculates the log prior probability of the model state

        The return value must be a **natural logarithm** of the probability.

        `MCMC` will short-circuit and abort the move if `probability` returns a
        negative infinity.  This can be used to avoid expensive likelihood
        calculations for obviously invalid moves, like going out of variable
        bounds.
        """
        ...


class Operator(Protocol):
    """
    Objects which propose moves by editing state

    It is the responsibility of the objects to track the parts of the state it
    might want to edit.  Typically these objects will be passed in the
    constructor.
    """

    weight: float
    """Influences the probability of the operator being picked

    On each step `MCMC` picks a random operator from the list passed to it.
    It uses this value to weight them.  So, the larger it is, the more
    often the operator will be picked, and visa versa.  This value is read
    once on startup.  Therefore, if it's changed mid-execution the old
    cached value will still be used.
    """

    def propose(self) -> Proposal:
        """Proposes a new MCMC step

        It is presumed that the operator will store all the references to
        parameters and trees it wants to edit and will change them accordingly.
        If a move cannot be proposed for any reason `Proposal.Reject` should be
        returned.  `MCMC` will deal with rolling back the state.
        """
        ...


class Callback(Protocol):
    """
    Custom callbacks

    `b3` supports arbitrary logging and checks via this protocol.  The `call`
    function is passed a reference to the main `MCMC` object, so the callback
    can either take state variables in the constructor of fetch them via the
    `MCMC` attributes.

    The `call` function won't be called after each step for efficiency.  See
    [`every`](#Callback.every) for configuring how often the callback will be
    invoked.
    """

    every: int
    """How often this callback should be called

    The `MCMC` will call each callback object when `index % every` is 0.  This
    value is read once when MCMC is created, so if it's changed during
    execution, the old `every` value will continue to be used.
    """

    def call(self, mcmc: MCMC) -> None:
        """
        A custom operation

        Used by loggers and other periodic actions.
        """


@runtime_checkable
class Stateful(Protocol):
    """
    Epoch-versioned objects for use with MCMC

    All stateful objects handled by `MCMC` must conform to this protocol.
    During each step operators can edit objects using whatever APIs provided.
    Then, at the end of the step, `MCMC` calls `accept` if the move has been
    accepted, or `reject` otherwise.
    """

    def accept(self) -> None:
        """Accept changes made during the current step"""

    def reject(self) -> None:
        """Reject changes made during the current step

        This method must roll the state of the object back to how it was at the
        beginning of the MCMC step.
        """



from __future__ import annotations

from collections.abc import Hashable, Iterator, Sequence
from dataclasses import dataclass
from datetime import timedelta
from typing import Optional, SupportsFloat

from ..b3 import Callback, Internal, Leaf, Node, Operator, Prior, Stateful, Tree
from ..b3.clocks import Clock
from ..b3.likelihoods import Likelihood
from ..b3.parameters import Scalable, Weights
from ..b3.substitutions import Substiution4
from ..data.msa import MSA
from ..data.newick import Tree as NewickTree
from ..rng import RNG
from ..stats.distributions import Sample

class Tree(Stateful):
    """A phylogenetic tree

    Unlike BEAST2, where the tree is implemented as a collection of nodes
    pointing to each other, in `b3` `Tree` is a self-contained data structure
    which holds all of the topology and heights.  This means that nodes
    (`Internal` and `Leaf`) are identifiers of nodes in a given `Tree` object,
    much like indices of an array.  This means that all operations, such as
    getting parents of node heights, have to go through `Tree`'s methods.

    The current implementation of `Tree` only supports bifurcating topologies.
    """

    def __init__(self, names: Sequence[str], rng: RNG):
        """
        - `names` is the list of names of leaf nodes.
        - `rng` is used to build a random tree.

        All heights are initially set to 0.  Use `set_height` for dated tips
        and `set_random_heights` to randomize the positions of internal nodes.
        """

    @classmethod
    def from_newick(_cls, newick: NewickTree) -> Tree:
        """
        Initializes the tree from a Newick object.

        The Newick tree must be strictly bifurcating and all of its edges must
        have a defined length.
        """

    def set_random_edges(self, rng: RNG):
        """
        Randomly sets edges between the nodes

        This methods creates a random [Prüfer sequence][wiki] and rearranges
        the graph according to it.  Note that it will always the internal node
        with the largest index (`num_nodes - 1`) the root.

        [wiki]: https://en.wikipedia.org/wiki/Pr%C3%BCfer_sequence
        """

    def set_random_heights(self, diff, rng: RNG):
        """
        Randomizes the heights of internal nodes

        Each internal node gets a height distributed uniformly between `diff`
        and `2 * diff` plus the height of the highest of its children.
        """

    @property
    def names(self) -> list[str]:
        """
        A list of all leaf names.

        The order is the same as leaf indices: the first name is that of
        `Leaf(0)`, the second one is `Leaf(1)`, and so on.
        """

    def scale(self, scale: float):
        """
        Multiplies the heights of all internal nodes by `scale`


        ### Exceptions

        Throws a `RuntimeError` if any of the internal nodes would be moved
        below either of its children.
        """

    def update_edge(self, edge: int, new_child: Node) -> None:
        """Sets the **child** of `edge` to `new_child`

        This will only change the child, so the parent (internal node from
        which `edge` comes out) will now have `node` as a child.

        This function doesn't do any validation, it's up to the operator to
        preserve the validity of the tree.
        """
    def set_height(self, node: Node, weigth: float) -> None:
        """Sets the height of `node` to `height`"""
    def set_root(self, node: Node) -> None:
        """Makes `node` the root of the tree

        As the topology can be temporarily broken while the edges are being
        swapped, `Tree` can't automatically figure out which node is the root
        one.  So, operators which change the root of the tree have to update it
        manually.
        """
    def swap_parents(self, a: Node, b: Node) -> None:
        """Swaps the parents of nodes `a` and `b`

        `a` and `b` must not be a descendant/ancestors and neither of them can
        be a root node.  If `a` and `b` share the same parent, they switch
        polarity (left child becomes the right child and visa versa).
        """
    @property
    def num_nodes(self) -> int:
        """The total number of nodes in the tree"""
    @property
    def num_internals(self) -> int:
        """The number of internal nodes (those with children)"""
    @property
    def num_leaves(self) -> int:
        """The number of leaf nodes"""
    @property
    def num_edges(self) -> int:
        """Total number of edges"""
    def is_internal(self, node: Node) -> bool:
        """Returns `True` if the node is internal"""
    def is_leaf(self, node: Node) -> bool:
        """Returns `True` if the node is a leaf"""
    def as_internal(self, node: Node) -> Optional[Internal]:
        """
        Converts `node` to the type `Internal` if it is internal, or returns
        `None` otherwise
        """
    def as_leaf(self, node: Node) -> Optional[Leaf]:
        """
        Converts `node` to the type `Leaf` if it is a leaf, or returns
        `None` otherwise
        """
    @property
    def root(self) -> Internal:
        """Returns the root node of the tree

        Note that the root node might change after tree has been edited, so the
        returned node is only guaranteed to be root as long as the tree hasn't
        been edited.
        """
    def height_of(self, node: Node) -> float:
        """Returns the height of `node`

        Height here means node's age in some unlabeled units.
        """
    def children_of(self, node: Internal) -> tuple[Node, Node]:
        """Returns a tuple of the left and right children of `node`

        This function takes the `Internal` type as its input, so it is
        guaranteed to always return the children.  See `as_internal` for
        converting general nodes to internal ones.
        """
    def other_child(self, parent: Internal, child: Node) -> Node:
        """Returns the child of `parent` other than `child`

        Throws an error if `child` isn't a child of `parent`.
        """
    def random_intersecting_edge(self, height: float, rng: RNG) -> Optional[int]:
        """Returns a random edge which intersects `height`

        "Intersects" here means that the edge parent is higher than `height`
        and the child is lower.  The comparisons are strict: if either node is
        exactly at `height`, the edge won't be picked.

        Returns `None` if there is no such node.
        """
    def edge_index(self, child: Node) -> int:
        """Returns the index of an edge from `child` to its parent"""
    def edge_length(self, edge: int) -> float:
        """Returns the length of `edge`

        The length is the distance between the parent and the child nodes of
        that edge.
        """
    def edge_nodes(self, edge: int) -> tuple[Node, Internal]:
        """Returns the `(child, parent)` tuple corresponding to an edge"""
    def parent_of(self, node: Node) -> Optional[Internal]:
        """Returns the parent of `node`, or `None` for the root node"""
    def is_grandparent(self, node: Internal) -> bool:
        """Returns `True` if both children of this node are also internal"""
    def num_grandparents(self) -> int:
        """Number of nodes for whom `is_grandparent` returns `True`"""
    def random_node(self, rng: RNG) -> Node:
        """Returns a random node from the tree

        It can be both an internal node or a leaf.  See `random_internal` and
        `random_leaf` for getting a random node of a specific kind.
        """
    def random_nonroot_node(self, rng: RNG) -> tuple[Internal, Internal]:
        """Returns a random non-root node and its parent"""
    def random_internal(self, rng: RNG) -> Internal:
        """Returns a random internal node"""
    def random_nonroot_internal(self, rng: RNG) -> tuple[Internal, Internal]:
        """Returns a random non-root internal and its parent"""
    def random_leaf(self, rng: RNG) -> Leaf:
        """Returns a random leaf node"""
    def nodes(self) -> Iterator[Node]:
        """An iterator over all of trees nodes

        All of the `Leaf` nodes go before `Internal` ones.
        """
    def leaf_by_name(self, name: str) -> Optional[Leaf]:
        """Gets a named leaf or `None` if the name is not found"""
    def internals(self) -> Iterator[Internal]:
        """An iterator over all of the trees internal nodes"""
    def leaves(self) -> Iterator[Leaf]:
        """An iterator over all of the trees leaf nodes"""
    def total_length(self) -> float:
        """The total length of all tree edges"""
    def has_dated_tips(self) -> bool:
        """Returns `True` if any of the leaves have non-0 height"""
    def validate(self) -> None:
        """Throws an exception if a tree is malformed

        This function ensures that:

        - No leaf has become anyone's parent.
        - All parent nodes are older than their children.
        - Parents match their children (mismatches can happen when
          `update_edge` is used incorrectly).
        - There's only one root (two or more can be set with `set_root`).
        - The tree is a tree, meaning that topologically it has no cycles and
          is connected.
        """
    def newick(self, internal_ids=False) -> str:
        """Returns the tree topology in the Newick format

        Leaf nodes will be labeled with the names passed to the constructor
        while the internal nodes are unlabeled.
        """

class Leaf(Hashable):
    """Leaf node of the phylogenetic tree

    Leaf nodes are the ones which are associated with a concrete sequence.
    Currently all leaf nodes have the distance of $0$, although that'll be
    subject to change in the future.
    """

class Internal(Hashable):
    """Internal anonymous node of the phylogenetic tree.

    Internals are the unnamed ancestors which form the tree.
    """

class Proposal:
    """A result of the move proposed by an operator

    While the operators edit the tree directly, they need to communicate the
    status of their move to `MCMC`.  This is the class used for that.
    """

    @classmethod
    def Reject(cls) -> Proposal:
        """Aborts the move unconditionally

        All of the trees and parameters are rolled back.  This is relatively
        fast, as it typically skips recalculating the likelihoods.
        """
    @classmethod
    def Hastings(cls, ratio: float) -> Proposal:
        """Proposes the move with the `ratio`

        This is the ratio from the Metropolis–Hastings algorithm.
        """
    @classmethod
    def Accept(cls) -> Proposal:
        """Accepts the move unconditionally"""

class CPU4Likelihood(Likelihood):
    """
    4-state DNA likelihood calculator.

    It's synchronous.  `Thread4Likelihood` and `CUDALikelihood` should be used
    for parallel calculations for alingments larger than 100Kb.
    """

    def __init__(
        self,
        msa: MSA,
        substitution: Substiution4,
        clock: Clock,
        tree: Tree,
    ): ...

class Thread4Likelihood(Likelihood):
    """
    Several `CPU4Likelihood`s in a trenchcoat.
    """

    def __init__(
        self,
        msa: MSA,
        substitution: Substiution4,
        clock: Clock,
        tree: Tree,
        *,
        thread_split_size: int = 400,
    ): ...

class CUDALikelihood(Likelihood):
    """
    Likelihood calculations on NVIDIA graphics cards.

    Only supports 4-state DNA models.

    `cuda_device` allows selecting the device index.
    """

    def __init__(
        self,
        msa: MSA,
        substitution: Substiution4,
        clock: Clock,
        tree: Tree,
        *,
        cuda_device: int = 0,
    ): ...

class MCMC:
    """
    The main object which runs the analysis
    """

    def __init__(
        self,
        burnin: int,
        length: int,
        state: Sequence[Stateful],
        priors: Sequence[Prior],
        operators: Sequence[Operator],
        likelihood: Likelihood,
        callbacks: Sequence[Callback],
        rng: RNG,
    ): ...
    @property
    def current_step(self) -> int:
        """
        Index of the current MCMC step

        Starts from 0, includes burn-in.
        """
    @property
    def state(self) -> list[Stateful]:
        """
        All stateful objects tracked by this `MCMC` instance
        """
    @property
    def priors(self) -> list[Prior]:
        """
        All priors
        """
    @property
    def operators(self) -> list[Operator]:
        """
        A list of all operator objects
        """
    @property
    def likelihood(self) -> Likelihood: ...
    @property
    def callbacks(self) -> list[Callback]:
        """
        All active callbacks
        """
    @property
    def rng(self) -> RNG:
        """
        Randomness source of this analysis

        This objects is passed to operators and used for internal randomness
        generation (such as picking the operator on each step).  Since the
        underlying object is shared, using it will alter the rest of the
        analysis.
        """
    @property
    def posterior(self) -> float:
        """Posterior probability for the last accepted step"""

    @property
    def prior(self) -> float:
        """Prior likelihood for the current step

        Note that unlike [`posterior`](#MCMC.posterior) and
        [`Likelihood`](#MCMC.Likelihood), this property isn't cached.  It will
        trigger a recalculation on all priors on each access.
        """

    @property
    def operator_statistics(
        self,
    ) -> list[tuple[Operator, list[int], timedelta, timedelta]]:
        """
        Operator statistics for this run

        Returns a list of `(operator, results, propose, likelihood)` tuples for
        each operator.  `propose` and `likelihood` records the total time the
        MCMC spent waiting for the operator to generate a proposal and
        calculate it respectively.  `operator` is the reference to the original
        operator object.  And `results` is a list of step results.
        """

    def run(self) -> None:
        """Start the simulation

        This yields flow control to the Rust core until the simulation is done.
        Press Ctrl+C to interrupt and stop the execution.
        """

    def measure_operator(self, operator_index: int, length: int) -> list[int]:
        """TODO"""

@dataclass
class EpochScale(Operator):
    """Scales a random epoch in a tree

    This parameter is analogous to BEAST2's `ScaleOperator` when it's used on a
    tree.  It will scale the full tree (so, for now, only its internal nodes,
    since leaves all have the height of 0).
    """

    tree: Tree
    factor: float
    """
    The scaling ratio will be sampled from `(factor, 1 / factor)`.  So, the
    factor must be between 0 and 1 and the smaller it is the larger the steps
    will be.
    """
    distribution: Sample[float]
    """Distribution from which the scale is sampled."""
    rng: RNG
    weight: float = 1

@dataclass(slots=True)
class SubtreeLeap(Operator):
    """
    Moves a node a distance, changing the topology randomly

    First, a distance delta is sampled from the distribution.  The operator
    selects a random node and all edges `delta` away from that node (down if
    the delta is negative or up and down if it's positive).  One of those edges
    is randomly selected and the node is spliced into it.  If the delta is
    above the root, the node will become the new root.
    """

    tree: Tree
    """The tree to edit."""
    distribution: Sample[float]
    """The distribution to draw the height move distance from"""
    rng: RNG
    weight: float = 1

@dataclass
class ConstantPopulation(Prior):
    """Constant population coalescent"""

    tree: Tree
    population: SupportsFloat

@dataclass
class ExponentialGrowth(Prior):
    tree: Tree
    population: SupportsFloat
    growth_rate: SupportsFloat

@dataclass
class Monophyly(Prior):
    """
    Ensures that a group of leaves form a monophyly

    Returns a static probability if the specified leaves are monophyletic or
    aborts the move otherwise.
    """

    tree: Tree
    leaves: Sequence[Leaf]

@dataclass
class Yule(Prior):
    """Uncalibrated Yule birth-rate model"""

    tree: Tree
    birth_rate: SupportsFloat

class Real(Stateful, Scalable):
    """
    Real multidimensional parameter

    It acts as a list of `float`s, except it implements
    [`Stateful`](#Stateful).  Note that this uses regular 64-bit floating point
    numbers underneath, not some custom arbitrary precision real type.
    """
    def __init__(self, *values: float): ...
    def __len__(self) -> int: ...
    def __getitem__(self, index: int) -> float: ...
    def __setitem__(self, index: int, value: float) -> None: ...
    def __float__(self) -> float: ...
    # richcmp
    def __lt__(self, other: float | Real) -> bool: ...
    def __le__(self, other: float | Real) -> bool: ...
    def __eq__(self, other) -> bool: ...
    def __ne__(self, other) -> bool: ...
    def __gt__(self, other: float | Real) -> bool: ...
    def __ge__(self, other: float | Real) -> bool: ...

class Integer(Stateful):
    """
    Integer multidimensional parameter

    It acts as a list of `int`s, except it implements [`Stateful`](#Stateful).
    """
    def __init__(self, *values: int): ...
    def __len__(self) -> int: ...
    def __getitem__(self, index: int) -> int: ...
    def __setitem__(self, index: int, value: int) -> None: ...
    def __int__(self) -> int: ...
    # richcmp
    def __lt__(self, other: int | Integer) -> bool: ...
    def __le__(self, other: int | Integer) -> bool: ...
    def __eq__(self, other) -> bool: ...
    def __ne__(self, other) -> bool: ...
    def __gt__(self, other: int | Integer) -> bool: ...
    def __ge__(self, other: int | Integer) -> bool: ...

class Boolean(Stateful):
    """
    Boolean multidimensional parameter

    It acts as a list of `bool`s, except it implements [`Stateful`](#Stateful).
    """

    def __init__(self, *values: bool): ...
    def __len__(self) -> int: ...
    def __getitem__(self, index: int) -> bool: ...
    def __setitem__(self, index: int, value: bool) -> None: ...
    # richcmp
    def __lt__(self, other: bool | Boolean) -> bool: ...
    def __le__(self, other: bool | Boolean) -> bool: ...
    def __eq__(self, other) -> bool: ...
    def __ne__(self, other) -> bool: ...
    def __gt__(self, other: bool | Boolean) -> bool: ...
    def __ge__(self, other: bool | Boolean) -> bool: ...

@dataclass
class JC:
    """
    Jukes-Cantor

    A simple model with equal state transition rates.

    Jukes and Cantor 1969, Evolution of Protein Molecules,
    <https://doi.org/10.1016/b978-1-4832-3211-9.50009-7>.
    """

@dataclass
class K80:
    """
    Kimura 80

    Equal base frequencies (A/C/G/T) with different transition (keeps
    purines/pyrimidines) and transversion (purine to pyrimidine and visa
    versa).

    Kimura 1980, A simple method for estimating evolutionary rates of base
    substitutions through comparative studies of nucleotide sequences,
    <https://doi.org/10.1007/BF01731581>.
    """

    kappa: SupportsFloat
    """
    A transition is taken to be kappa times more likely than a transversion.
    """

@dataclass
class HKY:
    """
    Hasegawa et al. 1985

    A model which can be thought of as a combination of K80 and F81: both base
    rates and transition/transversion ratio are configurable.

    Hasegawa et al. 1985, Dating of the human-ape splitting by a molecular
    clock of mitochondrial DNA, <https://doi.org/10.1007/BF02101694>.
    """

    frequencies: Weights | tuple[float, float, float, float]
    """
    DNA nucleotide frequencies

    Must have the length of 4, with each element corresponding to Adenine,
    Cytosine, Guanine, and Thymine respectively.
    """

    kappa: SupportsFloat
    """
    Transition/transversion ratio
    """



from dataclasses import dataclass
from math import log

from ...rng import RNG
from ...stats.distributions import Distribution
from .. import Internal, Node, Operator, Proposal, Tree
from ._util import assert_factor, sample_range


@dataclass(slots=True)
class RootSlide(Operator):
    """
    Scales the height of the root node

    The height will be scaled between `factor` and `1 / factor` sampled with
    `distribution`.  If the move will put the root below one of its children
    the operation gets rejected.
    """

    tree: Tree
    """The tree to edit."""
    factor: float
    distribution: Distribution
    """The distribution to draw the height move distance from"""
    rng: RNG
    weight: float = 1

    def __post_init__(self):
        assert_factor(self)

    def propose(self) -> Proposal:
        tree = self.tree
        rng = self.rng
        root = tree.root

        low, high = self.factor, 1 / self.factor
        scale = sample_range(low, high, self.distribution, rng)

        old_height = tree.height_of(root)

        left, right = tree.children_of(root)
        lower_height = max(tree.height_of(left), tree.height_of(right))

        new_height = (old_height - lower_height) * scale + lower_height

        tree.set_height(root, new_height)

        return Proposal.Hastings(-log(scale))



from typing import Optional

class RNG:
    """Random numbers generator.

    It's backed by a 64-bit output PCG, see the [Rust documentation][pcg] for
    details.

    It can be used standalone, but it was created as a Rust-native RNG which
    can be used efficiently by other Aspartik modules.

    [pcg]: https://docs.rs/rand_pcg/latest/rand_pcg/type.Pcg64.html
    """

    def __init__(self, seed: Optional[int] = None):
        """
        The seed is a positive integer less than $2^{64}$.  If no seed is
        passed the RNG will be seeded from the operating system data source.
        """
    def random_bool(self, ratio: float = 0.5) -> bool:
        """
        Rerturns `True` with the probability of `ratio`, which must be in the
        range $[0, 1]$.
        """
    def random_int(self, lower: int, upper: int) -> int:
        """Returns a random int in $[lower, upper)$."""
    def random_float(self, lower: float = 0.0, upper: float = 1.0) -> float:
        """Returns a float uniformly distributed on $[lower, upper]$

        See the [`rand` notes][float] for details on how the values are
        sampled.

        [float]: https://docs.rs/rand/latest/rand/distr/struct.StandardUniform.html#floating-point-implementation
        """



from dataclasses import dataclass
from math import log

from ...rng import RNG
from .. import Operator, Proposal, Tree


@dataclass(slots=True)
class WilsonBalding(Operator):
    """A version of a subtree regraft move

    Introduced in [this paper][paper], it picks a random subtree and inserts it
    in-between two other nodes.

    [paper]: https://doi.org/10.1093/genetics/161.3.1307
    """

    tree: Tree
    rng: RNG
    weight: float = 1

    def propose(self) -> Proposal:
        tree = self.tree
        rng = self.rng

        # pick a random non-root node
        while True:
            i_parent = tree.random_internal(rng)
            i_grandparent = tree.parent_of(i_parent)
            if i_grandparent is not None:
                break
        i_parent_height = tree.height_of(i_parent)

        i, i_brother = tree.children_of(i_parent)
        if rng.random_bool():
            i, i_brother = i_brother, i

        # Pick a node j_parent, such that it's above i_parent and one of its
        # children is below i_parent
        while True:
            j_parent = tree.random_internal(rng)
            j, j_brother = tree.children_of(j_parent)
            if rng.random_bool():
                j = j_brother

            if tree.height_of(j_parent) > i_parent_height > tree.height_of(j):
                break

        before = tree.height_of(i_grandparent) - max(
            tree.height_of(i), tree.height_of(i_brother)
        )
        after = tree.height_of(j_parent) - max(tree.height_of(i), tree.height_of(j))
        ratio = log(after / before)

        i_grandparent_to_i_parent = tree.edge_index(i_parent)
        j_parent_to_j = tree.edge_index(j)
        i_parent_to_i_brother = tree.edge_index(i_brother)

        # Cut out i_parent and replace it with a direct edge from grandparent
        # to i_brother
        tree.update_edge(i_grandparent_to_i_parent, i_brother)
        # Hook up i_parent to j_parent.  It's fine because we checked that
        # i_parent is lower than j_parent when selecting j
        tree.update_edge(j_parent_to_j, i_parent)
        # Replace i_brother edge from i_parent with j.  Once again, we've
        # enforced i_parent being above j earlier
        tree.update_edge(i_parent_to_i_brother, j)

        return Proposal.Hastings(ratio)



from ..._aspartik_rust_impl import _b3_rust_impl
from ._delta_exchange import DeltaExchange as DeltaExchange
from ._node_slide import NodeSlide as NodeSlide
from ._param_scale import ParamScale as ParamScale
from ._random_walk import RandomWalk as RandomWalk
from ._root import RootSlide as RootSlide
from ._spr import SubtreePruneRegraft as SubtreePruneRegraft
from ._subtree_slide import SubtreeSlide as SubtreeSlide
from ._tree_exchange import (
    BeastNarrowExchange as BeastNarrowExchange,
    BeastWideExchange as BeastWideExchange,
    NarrowExchange as NarrowExchange,
    WideExchange as WideExchange,
)
from ._tree_scale import TreeScale
from ._updown import UpDown
from ._wilson_balding import WilsonBalding as WilsonBalding

EpochScale = _b3_rust_impl.EpochScale
SubtreeLeap = _b3_rust_impl.SubtreeLeap


__all__ = [
    "DeltaExchange",
    "NodeSlide",
    "ParamScale",
    "RandomWalk",
    "RootSlide",
    "SubtreePruneRegraft",
    "SubtreeSlide",
    "BeastNarrowExchange",
    "BeastWideExchange",
    "NarrowExchange",
    "WideExchange",
    "TreeScale",
    "UpDown",
    "WilsonBalding",
    # Rust
    "EpochScale",
    "SubtreeLeap",
]



from .._aspartik_rust_impl._math_rust_impl import (
    binomial as binomial,
    digamma as digamma,
    digamma_inv as digamma_inv,
    # exponential
    ei as ei,
    # erf
    erf as erf,
    erf_inv as erf_inv,
    erfc as erfc,
    erfc_inv as erfc_inv,
    # factorial
    factorial as factorial,
    # gamma
    gamma as gamma,
    gamma_li as gamma_li,
    gamma_lr as gamma_lr,
    gamma_ui as gamma_ui,
    gamma_ur as gamma_ur,
    generalized_harmonic as generalized_harmonic,
    # harmonic
    harmonic as harmonic,
    ln_binomial as ln_binomial,
    ln_factorial as ln_factorial,
    ln_gamma as ln_gamma,
    # logistic
    logistic as logistic,
    logit as logit,
)



"""Classes which record the state of the simulation.

All classes here adhere to the `Callback` protocol and should be passed as
such.
"""

import json
from collections.abc import Callable, Mapping
from dataclasses import dataclass, field
from io import TextIOBase
from typing import Any

from . import MCMC, Callback, Prior, Tree
from .parameters import Real, Weights


@dataclass(slots=True)
class TreeLogger(Callback):
    """Records the topology of the tree into a `.trees` file."""

    tree: Tree
    path: str
    """
    Path to the file where the trees will be appended in Newick format, one per
    line.  It's opened verbatim (the `.trees` extension won't be added).
    """
    every: int
    """How often the logger will be called"""

    def __post_init__(self):
        self._file = open(self.path, "w")

    def log(self, mcmc: MCMC):
        line = self.tree.newick()
        self._file.write(line)
        self._file.write("\n")

    def __getstate__(self):
        # None ignores __dict__ which contains the file handle
        return (None, self.__slots__)


@dataclass(slots=True)
class PrintLogger(Callback):
    every: int

    def __post_init__(self):
        print(f"{'step':>16}{'posterior':>16}{'likelihood':>16}{'prior':>16}")

    def log(self, mcmc: MCMC):
        likelihood = mcmc.likelihood.likelihood()
        print(
            f"{mcmc.current_step:>16}{mcmc.posterior:>16.2f}{likelihood:>16.2f}{mcmc.prior:>16.2f}"
        )


def _serialize(item):
    if isinstance(item, Real):
        return float(item)

    elif isinstance(item, Weights):
        return list(item)

    elif isinstance(item, Prior):
        return item.probability()

    elif callable(item):
        return item()


@dataclass(slots=True)
class ValueLogger(Callback):
    items: Mapping[str, Any]
    path: str
    every: int

    _file: TextIOBase = field(init=False)

    def __post_init__(self):
        self._file = open(self.path, "w")

    def log(self, mcmc: MCMC):
        entry_json = json.dumps(self.items, default=_serialize)
        self._file.write(entry_json)
        self._file.write("\n")
        self._file.flush()

    def __getstate__(self):
        return (None, self.__slots__)



from typing import Optional, Protocol, runtime_checkable

from .._aspartik_rust_impl import _stats_rust_impl
from ..rng import RNG


@runtime_checkable
class Continuous[T](Protocol):
    def pdf(self, x: T) -> float: ...
    def ln_pdf(self, x: T) -> float: ...


@runtime_checkable
class ContinuousCDF[T](Continuous[T], Protocol):
    def cdf(self, x: T) -> float: ...
    def sf(self, x: T) -> float: ...
    def inverse_cdf(self, p: float) -> T: ...
    @property
    def lower(self) -> T: ...
    @property
    def upper(self) -> T: ...


@runtime_checkable
class Discrete[T](Protocol):
    def pmf(self, x: T) -> float: ...
    def ln_pmf(self, x: T) -> float: ...


@runtime_checkable
class DiscreteCDF[T](Discrete[T], Protocol):
    def cdf(self, x: T) -> float: ...
    def sf(self, x: T) -> float: ...
    def inverse_cdf(self, p: float) -> T: ...
    @property
    def lower(self) -> T: ...
    @property
    def upper(self) -> T: ...


@runtime_checkable
class Distribution(Protocol):
    def mean(self) -> Optional[float]: ...
    def median(self) -> Optional[float]: ...
    def variance(self) -> Optional[float]: ...
    def std_dev(self) -> Optional[float]: ...
    def entropy(self) -> Optional[float]: ...
    def skewness(self) -> Optional[float]: ...


@runtime_checkable
class Sample[T](Protocol):
    def sample(self, rng: RNG) -> T: ...


# ruff: noqa: F822
__all__ = [
    # Classes
    "Beta",
    "BetaError",
    "Exp",
    "ExpError",
    "Gamma",
    "GammaError",
    "InverseGamma",
    "InverseGammaError",
    "Laplace",
    "LaplaceError",
    "LogNormal",
    "LogNormalError",
    "Normal",
    "NormalError",
    "Poisson",
    "PoissonError",
    "Uniform",
    "UniformError",
    # protocols
    "Continuous",
    "ContinuousCDF",
    "Discrete",
    "DiscreteCDF",
    "Distribution",
    "Sample",
]

for item in __all__[:18]:
    locals()[item] = getattr(_stats_rust_impl.distributions, item)


def __dir__():
    return __all__



def is_close(
    given: float,
    expected: float,
    epsilon=2.2204460492503131e-16,
    relative=1e-16,
    ulps=4,
) -> bool: ...

# float
def sign(x: float) -> bool: ...
def exponent(x: float) -> int: ...
def exponent_bits(x: float) -> int: ...
def mantissa(x: float) -> int: ...
def mantissa_bits(x: float) -> int: ...

# functions
def erf(x: float) -> float: ...
def erf_inv(x: float) -> float: ...
def erfc(x: float) -> float: ...
def erfc_inv(x: float) -> float: ...
def ei(n: int, x: float) -> float: ...
def factorial(x: float) -> float: ...
def ln_factorial(x: float) -> float: ...
def binomial(n: int, k: int) -> float: ...
def ln_binomial(n: int, k: int) -> float: ...
def gamma(x: float) -> float: ...
def ln_gamma(x: float) -> float: ...
def gamma_ui(s: float, x: float) -> float: ...
def gamma_li(s: float, x: float) -> float: ...
def gamma_ur(s: float, x: float) -> float: ...
def gamma_lr(s: float, x: float) -> float: ...
def digamma(x: float) -> float: ...
def digamma_inv(x: float) -> float: ...
def harmonic(t: int) -> float: ...
def generalized_harmonic(n: int, m: float) -> float: ...
def logistic(p: float) -> float: ...
def logit(p: float) -> float: ...



from dataclasses import dataclass
from math import log

from ...rng import RNG
from ...stats.distributions import Sample
from .. import Internal, Node, Operator, Proposal, Tree
from ._util import family


@dataclass(slots=True)
class SubtreeSlide(Operator):
    """"""

    tree: Tree
    """The tree to edit."""
    distribution: Sample[float]
    """
    The distribution which will sample the new node height on the interval
    between its parent and the closest child.
    """
    rng: RNG
    weight: float = 1

    def propose(self) -> Proposal:
        """
        If there are no non-root internal nodes, the operator will bail with
        `Proposal.Reject`.
        """

        rng = self.rng
        tree = self.tree
        root = tree.root

        ratio = 0.0

        # automatically fail on trees without non-root internal nodes
        if tree.num_internals == 1:
            return Proposal.Reject()

        node, sibling, parent, grandparent = family(tree, rng)
        delta = self.distribution.sample(rng)
        old_height = tree.height_of(parent)
        new_height = old_height + delta

        if delta > 0:
            if grandparent is not None and tree.height_of(grandparent) < new_height:
                # two nodes whose edge intersects `new_height`
                new_child = parent
                new_parent = grandparent

                while tree.height_of(new_parent) < new_height:
                    new_child = new_parent
                    new_parent = tree.parent_of(new_child)
                    if new_parent is None:
                        break

                parent_to_sibling = tree.edge_index(sibling)
                grandparent_to_parent = tree.edge_index(parent)

                if new_parent is None:
                    # new_child was the root
                    tree.update_edge(parent_to_sibling, new_child)
                    tree.update_edge(grandparent_to_parent, sibling)
                    tree.set_root(parent)
                else:
                    new_parent_to_new_child = tree.edge_index(new_child)

                    tree.update_edge(parent_to_sibling, new_child)
                    tree.update_edge(grandparent_to_parent, sibling)
                    tree.update_edge(new_parent_to_new_child, parent)

                num_reverse_sources = len(intersections(tree, new_child, old_height))
                ratio = -log(num_reverse_sources)
        else:
            if tree.height_of(node) > new_height:
                return Proposal.Reject()

            if tree.height_of(sibling) > new_height:
                # topological changes

                destinations = intersections(tree, sibling, new_height)
                random_idx = rng.random_int(0, len(destinations))
                new_child = destinations[random_idx]
                new_parent = tree.parent_of(new_child)

                parent_to_sibling = tree.edge_index(sibling)
                new_parent_to_new_child = tree.edge_index(new_child)

                if parent == tree.root:
                    tree.update_edge(parent_to_sibling, new_child)
                    tree.update_edge(new_parent_to_new_child, parent)

                    tree.set_root(sibling)
                else:
                    grandparent_to_parent = tree.edge_index(parent)

                    tree.update_edge(parent_to_sibling, new_child)
                    tree.update_edge(grandparent_to_parent, sibling)
                    tree.update_edge(new_parent_to_new_child, parent)

                ratio = -log(len(destinations))

        tree.set_height(parent, new_height)

        tree.validate()

        return Proposal.Hastings(ratio)


def intersections(tree: Tree, node: Node, height: float) -> list[Node]:
    if tree.height_of(node) < height:
        return [node]

    if isinstance(node, Internal):
        left, right = tree.children_of(node)
        return intersections(tree, left, height) + intersections(tree, right, height)
    else:
        return []



from __future__ import annotations

from typing import BinaryIO, TextIO

from ..data.fasta import DNARecord

class FastaReader:
    def __init__(self, reader: TextIO | BinaryIO): ...
    @classmethod
    def from_file(cls, path: str) -> FastaReader: ...
    def __iter__(self) -> FastaReader: ...
    def __next__(self) -> DNARecord: ...



from dataclasses import dataclass, field
from typing import SupportsFloat

from ...stats.distributions import Continuous, Discrete
from .. import Prior
from ..parameters import Real


@dataclass(slots=True)
class Distribution(Prior):
    """
    Calculates prior probability of a single-dimensional parameter according to
    a distribution.
    """

    param: SupportsFloat
    distribution: Continuous
    """Distribution against which the parameter prior is calculated."""

    def probability(self) -> float:
        return self.distribution.ln_pdf(float(self.param))



from typing import Optional, Protocol, Self, TypeVar, runtime_checkable

from ..rng import RNG

@runtime_checkable
class Continuous(Protocol):
    def pdf(self, x: float) -> float: ...
    def ln_pdf(self, x: float) -> float: ...

@runtime_checkable
class ContinuousCDF(Continuous, Protocol):
    def cdf(self, x: float) -> float: ...
    def sf(self, x: float) -> float: ...
    def inverse_cdf(self, p: float) -> float: ...
    @property
    def lower(self) -> float: ...
    @property
    def upper(self) -> float: ...

@runtime_checkable
class Discrete(Protocol):
    def pmf(self, x: int) -> float: ...
    def ln_pmf(self, x: int) -> float: ...

@runtime_checkable
class DiscreteCDF(Discrete, Protocol):
    def cdf(self, x: int) -> float: ...
    def sf(self, x: int) -> float: ...
    def inverse_cdf(self, p: float) -> int: ...
    @property
    def lower(self) -> int: ...
    @property
    def upper(self) -> int: ...

@runtime_checkable
class Distribution(Protocol):
    def mean(self) -> Optional[float]: ...
    def median(self) -> Optional[float]: ...
    def variance(self) -> Optional[float]: ...
    def std_dev(self) -> Optional[float]: ...
    def entropy(self) -> Optional[float]: ...
    def skewness(self) -> Optional[float]: ...

@runtime_checkable
class Sample[T](Protocol):
    def sample(self, rng: RNG) -> T: ...

# Concrete distributions

class Beta(ContinuousCDF, Continuous, Distribution, Sample):
    def __init__(self, shape_a: float, shape_b: float): ...
    @property
    def shape_a(self) -> float: ...
    @property
    def shape_b(self) -> float: ...

class BetaError:
    InvalidAlpha: Self
    InvalidBeta: Self

class Exp(ContinuousCDF, Continuous, Distribution, Sample[float]):
    def __init__(self, rate: float): ...
    @property
    def rate(self) -> float: ...

class ExpError:
    RateInvalid: Self

class Gamma(ContinuousCDF, Continuous, Distribution, Sample[float]):
    def __init__(self, shape: float, rate: float): ...
    @property
    def shape(self) -> float: ...
    @property
    def rate(self) -> float: ...

class GammaError:
    ShapeInvalid: Self
    RateInvalid: Self
    ShapeAndRateInfinite: Self

class InverseGamma(ContinuousCDF, Continuous, Distribution, Sample[float]):
    def __init__(self, shape: float, rate: float): ...
    @property
    def shape(self) -> float: ...
    @property
    def rate(self) -> float: ...

class InverseGammaError:
    ShapeInvalid: Self
    RateInvalid: Self

class Laplace(ContinuousCDF, Continuous, Distribution, Sample[float]):
    def __init__(self, location: float, scale: float): ...
    @property
    def location(self) -> float: ...
    @property
    def scale(self) -> float: ...

class LaplaceError:
    LocationInvalid: Self
    ScaleInvalid: Self

class LogNormal(ContinuousCDF, Continuous, Distribution, Sample[float]):
    def __init__(self, location: float, scale: float): ...
    @property
    def location(self) -> float: ...
    @property
    def scale(self) -> float: ...

class LogNormalError:
    LocationInvalid: Self
    ScaleInvalid: Self

class Normal(ContinuousCDF, Continuous, Distribution, Sample[float]):
    def __init__(self, mean: float, std_dev: float): ...

class NormalError:
    MeanInvalid: Self
    StandardDeviationInvalid: Self

class Poisson(DiscreteCDF, Discrete, Distribution, Sample[int]):
    def __init__(self, lambda_: float): ...
    @property
    def lambda_(self) -> float: ...

class PoissonError:
    LambdaInvalid: Self

class Uniform(ContinuousCDF, Continuous, Distribution, Sample[float]):
    def __init__(self, min: float, max: float): ...
    @property
    def min(self) -> float: ...
    @property
    def max(self) -> float: ...

class UniformError:
    MinInvalid: Self
    MaxInvalid: Self
    MaxNotGreaterThanMin: Self



from collections.abc import Sequence
from copy import deepcopy
from dataclasses import dataclass
from typing import Protocol, SupportsFloat, runtime_checkable

from . import Stateful, Tree


@runtime_checkable
class Scalable(Protocol):
    def scale(self, factor: float) -> int:
        """
        Scales all values of a parameter and returns the number of dimensions.
        """

        ...


class Real(Stateful, SupportsFloat, Scalable):
    __slots__ = ("_value", "_backup")
    _value: float
    _backup: float

    def __init__(self, value: SupportsFloat):
        self._value = float(value)
        self._backup = self._value

    # comparison
    def __lt__(self, other: SupportsFloat) -> bool:
        return self._value < float(other)

    def __le__(self, other: SupportsFloat) -> bool:
        return self._value <= float(other)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, SupportsFloat):
            return self._value == float(other)
        else:
            return False

    def __ne__(self, other: object) -> bool:
        if isinstance(other, SupportsFloat):
            return self._value != float(other)
        else:
            return False

    def __gt__(self, other: SupportsFloat) -> bool:
        return self._value > float(other)

    def __ge__(self, other: SupportsFloat) -> bool:
        return self._value >= float(other)

    # math
    def __iadd__(self, other: SupportsFloat):
        self._value += float(other)
        return self

    def __isub__(self, other: SupportsFloat):
        self._value -= float(other)
        return self

    def __imul__(self, other: SupportsFloat):
        self._value *= float(other)
        return self

    def __idiv__(self, other: SupportsFloat):
        self._value /= float(other)
        return self

    def __repr__(self) -> str:
        return repr(self._value)

    def __float__(self) -> float:
        return self._value

    def set(self, value: SupportsFloat) -> None:
        self._value = float(value)

    def scale(self, factor: float) -> int:
        self *= factor
        return 1

    def accept(self):
        self._backup = self._value

    def reject(self):
        self._value = self._backup


class Weights(Stateful, Scalable):
    __slots__ = ("_value", "_backup", "_length")
    _value: list[float]
    _backup: list[float]
    _length: int

    def __init__(self, *args: float):
        self._value = list(args)
        self._backup = deepcopy(self._value)
        self._length = len(self._value)

    def __len__(self) -> int:
        return self._length

    def __getitem__(self, i: int) -> float:
        return self._value[i]

    def __setitem__(self, i: int, value: float):
        self._value[i] = value

    def __iter__(self):
        return iter(self._value)

    def __repr__(self) -> str:
        return repr(self._value)

    # comparison
    # XXX: deduplicate?
    def __lt__(self, other: SupportsFloat) -> bool:
        other = float(other)
        for item in self:
            if not item < other:
                return False
        return True

    def __le__(self, other: SupportsFloat) -> bool:
        other = float(other)
        for item in self:
            if not item <= other:
                return False
        return True

    def __gt__(self, other: SupportsFloat) -> bool:
        other = float(other)
        for item in self:
            if not item > other:
                return False
        return True

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Weights):
            return self._value == other._value
        else:
            return False

    def __ne__(self, other: object) -> bool:
        if isinstance(other, Weights):
            return self._value != other._value
        else:
            return False

    def __ge__(self, other: SupportsFloat) -> bool:
        other = float(other)
        for item in self:
            if not item >= other:
                return False
        return True

    def scale(self, factor: float) -> int:
        for i in range(len(self)):
            self[i] *= factor
        return len(self)

    def accept(self):
        for i in range(len(self)):
            self._backup[i] = self._value[i]

    def reject(self):
        for i in range(len(self)):
            self._value[i] = self._backup[i]


@dataclass
class Internals(Scalable):
    tree: Tree

    def scale(self, factor: float) -> int:
        self.tree.scale(factor)

        return self.tree.num_internals


@dataclass
class Root:
    tree: Tree

    def __float__(self) -> float:
        return self.tree.height_of(self.tree.root)



from dataclasses import dataclass
from math import log

from ...rng import RNG
from ...stats.distributions import Distribution
from .. import Operator, Proposal
from ..parameters import Real, Scalable
from ._util import assert_factor, sample_range


@dataclass(slots=True)
class UpDown(Operator):
    """
    TODO
    """

    up: Scalable
    """The parameter to scale up."""
    down: Scalable
    """The parameter to scale down."""
    factor: float
    """
    The scale ratio will be sampled from `(factor, 1 / factor)`.  So, the
    smaller the factor, the larger the moves proposed by this operator are.
    This also means that `factor` must be within `(0, 1)`.
    """
    distribution: Distribution
    """The distribution from which to sample the scaling factor."""
    rng: RNG
    weight: float = 1

    def __post_init__(self):
        assert_factor(self)

    def propose(self) -> Proposal:
        low, high = self.factor, 1 / self.factor
        scale = sample_range(low, high, self.distribution, self.rng)

        try:
            num_scaling_up = self.up.scale(scale)
            num_scaling_down = self.down.scale(1 / scale)
        except:
            return Proposal.Reject()

        ratio = log(scale) * (num_scaling_up - num_scaling_down - 2)
        return Proposal.Hastings(ratio)
